#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "main.h"
#include "hash.h"


/* Search the given string 'haystack' for either tag and return a char * 
 * to that instnace.
 */
static const char *locate_tag(
    const char *haystack,
    const char *tag,
    const char *tag_alt)
{
    const char *st;
    const char *safe;

    st = safe = haystack;
    if (!(st = strstr(st, tag)))
      if (!(st = strstr(safe, tag_alt)))
        return NULL;

    return st;
}


static rss_t *raw_to_rss(const char *buf, int *n_entries)
{
    int         i, n_rss;
    size_t      len;
    const char *st;
    const char *en;
    const char *item_st;
    const char *item_en;
    rss_t      *rss;

    if (n_entries)
      *n_entries = 0;
   
    /* RSS start */ 
    st = buf;
    if (!(st = locate_tag(st, "<item>", "<ITEM>")))
      return NULL;

    buf = st;

    /* Count entries */
    n_rss = 0;
    while (st)
    {
        if (!(st = locate_tag(st, "<item>", "<ITEM>")))
          break;
        
        ++n_rss;
        ++st;
    }

    rss = malloc(sizeof(rss_t) * n_rss);
    memset(rss, 0, sizeof(rss_t) * n_rss);
    
    /* Look for title and link elements */
    st = item_st = buf;
    for (i=0; i<n_rss; ++i)
    { 
        /* Locate start/end of item */
        if (!(item_st = locate_tag(item_st, "<item>", "<ITEM>")))
          break;
        
        if (!(item_en = locate_tag(item_st, "</item>", "</ITEM>")))
          break;

        ++item_st;
        ++item_en;
        
        /* Get title */
        st = item_st;
        if (!(st = locate_tag(st, "<title>", "<TITLE>")))
          break;

        if (!(en = locate_tag(st, "</title>", "</TITLE>")))
          break;

        if ((st < item_en) && (en < item_en))
        {
            st += strlen("<title>");
            len = MIN((en-st), (RSS_TITLE_LENGTH-1));
            strncpy(rss[i].title, st, len);
            ++st;
        }

        /* Get URL */
        st = item_st;
        if (!(st = locate_tag(st, "<link>", "<LINK>")))
          continue;

        if (!(en = locate_tag(st, "</link>", "</LINK>")))
          break;

        if ((st < item_en) && (en < item_en))
        {
            st += strlen("<link>");
            len = MIN((en-st), (RSS_URL_LENGTH-1));
            strncpy(rss[i].url, st, len);
            ++st;
        }
    }

    if (n_entries)
      *n_entries = n_rss;

    return rss;
}


rss_t *utils_get_rss(const char *url, int *n_entries)
{
    int                 sd, buf_off;
    ssize_t             buf_sz, n_bytes;
    char               *c, *path, *buf;
    char                hostname[256], data[BLOCK_SIZE], query[256];
    struct hostent     *host;
    struct sockaddr_in  addr;

    /* Ignore protocol and get just the hostname */
    if ((c = strstr(url, "://")))
      strncpy(hostname, c + strlen("://"), sizeof(hostname));
    else
      strncpy(hostname, url, sizeof(hostname));
    
    /* Remove path */
    if ((path = strchr(hostname, '/')))
    {
        *path = '\0';
        path += 1;
    }

    /* Get host address */
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(80);
    ERR_NULL((host = gethostbyname(hostname)))
    memcpy(&addr.sin_addr, host->h_addr_list[0], host->h_length);

    /* Create socket and connect */
    ERR_NEG(sd = socket(AF_INET, SOCK_STREAM, 0))
    ERR_NEG(connect(sd, (const struct sockaddr *)&addr, sizeof(addr)))

    /* Query */
    memset(query, 0, sizeof(query));
    snprintf(query, sizeof(query), "GET /%s HTTP/1.1\r\nHost: %s\r\n\r\n",
             path, hostname);
    ERR_NEG(send(sd, query, 256, 0))

    /* Get data */
    buf = malloc(BLOCK_SIZE);
    buf_sz = BLOCK_SIZE;
    buf_off = 0;
    while ((n_bytes = recv(sd, data, sizeof(data), 0)))
    {
        if ((buf_off + n_bytes) > buf_sz)
        {
            buf_sz += BLOCK_SIZE;
            buf = realloc(buf, buf_sz);
        }

        memcpy(buf+buf_off, data, n_bytes);
        buf_off += n_bytes;
    }

    return raw_to_rss(buf, n_entries);
}


hash_t *utils_get_lyrics(const char *fname, int *n_keys)
{
    int     idx;
    char    c;
    FILE   *fp;
    hash_t *hash;
    char   *w, word[32] = {0};

    ERR_NULL(fp = fopen(fname, "r"))
    hash = hash_create(HASH_DEFAULT_SIZE);
    idx = 0;

    if (n_keys)
      *n_keys = HASH_DEFAULT_SIZE;

    while ((c = fgetc(fp)) != EOF)
    {   
        if (islower(c) || isupper(c) || (c == '-') || (c == '\''))
          word[idx++] = tolower(c);
        else if ((ispunct(c) || isspace(c)) && (idx > 1))
        {
            w = calloc(1, MIN(idx, 32));
            strncpy(w, word, MIN(idx, (32-1)));
            hash_add(hash, w, MIN(idx+1, 32),
                     hash_default_hash(w, strlen(w) + 1));
            idx = 0;
            memset(word, 0, MIN(idx, 32));
        }
    }

    return hash;
}
